/*-----------------------------------------------------------------------------
 * CS 155 - Computer & Network Security - Spring 2005
 *
 * file:  ids.c
 * Contact: stinson@stanford
 *          sguha05@stanford
 *
 * Description: use pcap to gather data which we'll use to determine
 *              whether a port scan or a TCP SYN (DoS) attack is under way
 *---------------------------------------------------------------------------*/

#include "ids.h"

/* ---------------------------------------------------------
 *
 *
 * 
 *
 * --------------------------------------------------------- */
int main(int argc, char *argv[]) {

  if (getuid() != 0) {
    fprintf(stderr, "This program must be run as root!\n");
    exit(-1); }
  
  /* ----- the # of TCP SYNs which constitutes an attack ----- */
  int tcp_syn_thresh = -1;  

  /* ----- time over which those SYNs must be recv'd ----- */
  int tcp_syn_secs = -1;    
  
  /* ----- the # port scan probe pkts which constitutes an attack ----- */
  int port_scan_thresh = -1; 

  /* ----- time over which those pkts must be recv'd ----- */
  int port_scan_secs = -1;   
  
  /* ----- name of file to which to log tcp syn flood info ----- */
  char *tcp_syn_file = 0;    

  /* ----- name of file to which to log port scan info ----- */
  char *port_scan_file = 0;
  
  int c;
  
  /* ------------- parse command-line options ------------- */
  while ((c = getopt(argc, argv, "ht:s:p:S:f:F:")) != EOF)  {
    
    switch (c) {
      
    case 'h': /* ----- to get help menu (list of options) ----- */
      usage(argv[0]);
      exit(0);
      break;
      
    case 't': /* ----- tcp syn threshold ----- */
      tcp_syn_thresh = atoi( (char*) optarg );
      break;
      
    case 's': /* ----- tcp syn seconds ----- */
      tcp_syn_secs = atoi( (char *) optarg );
      break;
      
    case 'p': /* ----- port scan threshold ----- */
      port_scan_thresh = atoi( (char *) optarg );
      break;
      
    case 'S': /* ----- port scan seconds ----- */
      port_scan_secs = atoi( (char*) optarg );
      break;
      
    case 'f': /* ----- file to log tcp syn flood ----- */
      if ( strlen(optarg) > 0 ) {
	tcp_syn_file = (char*)malloc( strlen(optarg) + 1 );
	memcpy( tcp_syn_file, optarg, strlen(optarg) );
	memset( tcp_syn_file + strlen(optarg), '\0', 1 );
      }
      break;
      
    case 'F': /* ----- file to log port scan ----- */
      if ( strlen(optarg) > 0 ) {
	port_scan_file = (char*)malloc( strlen(optarg) + 1 );
	memcpy( port_scan_file, optarg, strlen(optarg) );
	memset( port_scan_file + strlen(optarg), '\0', 1 );
      }
      break;
      
    } /* switch */
  } /* -- while -- */
  
  int retval = monitor( tcp_syn_thresh, 
			tcp_syn_secs, 
			port_scan_thresh, 
			port_scan_secs,
			tcp_syn_file, 
			port_scan_file );
  
  if ( tcp_syn_file )
    free( tcp_syn_file );
  
  if ( port_scan_file )
    free( port_scan_file );

  exit(retval);
  
  return retval;
}

/* ----------------------------------------------------------------
 * tcp_syn_thresh : # of TCP SYNs that constitute an attack
 *
 * tcp_syn_secs   : max # of secs over which those SYNs must arrive
 *                  in order to constitute an attack
 * 
 * port_scan_thresh : # of probes in order to constitute an attack
 *
 * port_scan_secs   : max # of secs over which those port probes
 *                    must arrive in order to constitute an attack
 *
 * IMPLIED: 
 *
 *  -- tcp_syn_thresh SYNs must arrive destined for a single IP
 *     AND a single port over at most tcp_syn_secs in order to be 
 *     considered a TCP SYN attack (and we only count SYNs to
 *     listening ports ... i.e. those for which SYN-ACKs are sent)
 *
 *  -- port_scan_thresh probes[*] must arrive for ANY destination
 *     IP on the LAN within port_scan_secs in order to be
 *     considered a PORT SCAN 
 *
 * [*]port scan probes : 
 *
 *   - anything which could learn whether a TCP port on one of
 *     the destination IPs is listening or not
 *
 *   - also ICMP echo requests (pings)
 * ---------------------------------------------------------------- */
int monitor( int tcp_syn_thresh,   
	     int tcp_syn_secs, 
	     int port_scan_thresh, 
	     int port_scan_secs,
	     char *tcp_syn_file,   
	     char *port_scan_file ) {
  
  /* --------------------------------------------------------------- */
  /* ++++++++++++++++ set up pscan stuff + headers +++++++++++++++++ */
  /* --------------------------------------------------------------- */
  pcap_t                    *handle;          /* Session handle */
  char 		            *dev;             /* The device to sniff on */
  char 		            errbuf[PCAP_ERRBUF_SIZE]; /*Err string */
  bpf_u_int32               mask;             /* Our netmask */
  bpf_u_int32               net;              /* Our IP */
  struct pcap_pkthdr        header;           /* Header that pcap gives us */
  const u_char              *packet;          /* The actual packet */
  
  const struct ether_header *ether_header;    /* The Ethernet header */
  const struct ip           *ip;              /* The IP header */
  const struct tcphdr       *tcphdr;          /* The TCP header */
  const struct udphdr       *udphdr;          /* The UDP header */
  const struct icmphdr      *icmp;            /* The ICMP header */
  const char                *payload;         /* Packet payload */
  
  /* Used for debugging purposes */
  char                      *proto;
  u_short                   src_port;
  u_short                   dst_port;
  
  /* Sizes of some relevant headers */
  int size_ethernet       = sizeof(struct ether_header); 
  int size_ip             = sizeof(struct ip);
  int size_tcp            = sizeof(struct tcphdr);
  int size_icmp           = sizeof(struct icmphdr);

  int detect_syn_flood = 0, detect_port_scan = 0;
  
  /* ----------------------------------------------------------
   * don't detect tcp syn floods
   * ---------------------------------------------------------- */
  if ( ( tcp_syn_thresh == -1 ) || ( tcp_syn_secs == -1 ) ) {

#ifdef DEBUG
    fprintf(stderr,"Not detecting tcp syn floods...\n");
#endif
  }

  /* ----------------------------------------------------------
   * do detect tcp syn floods
   * ---------------------------------------------------------- */
  else {

    detect_syn_flood = 1;
    
#ifdef DEBUG
    fprintf(stderr,"Detecting tcp syn floods...\n");
#endif    

    /* ----------------------------------------------------------
     * initialize TCP SYN flood detection data structures
     * ---------------------------------------------------------- */

  }
  
  /* ------------------------------------------------------------
   * don't detect port scans
   * ------------------------------------------------------------ */
  if ( ( port_scan_thresh == -1 ) || ( port_scan_secs == -1 ) ) {

#ifdef DEBUG
    fprintf(stderr,"Not detecting port scans...\n");
#endif
    
  }
  
  /* ------------------------------------------------------------
   * do detect port scans
   * ------------------------------------------------------------ */
  else {

    detect_port_scan = 1;

#ifdef DEBUG
    fprintf(stderr,"Detecting port scans...\n");
#endif    

    /* ----------------------------------------------------------
     * initialize port scan detection data structures 
     * ---------------------------------------------------------- */
  }
  
  /* ------------------------------------------------------------
   * don't detect either so exit...................
   * ------------------------------------------------------------ */
  if ( !detect_syn_flood && !detect_port_scan ) {
    fprintf(stderr,"Not detecting SYN flood or port scan so done...\n");
    return 0; }

  /* ------------------------------------------------------------
   * detecting both so exit.......................
   * ------------------------------------------------------------ */
  if ( detect_syn_flood && detect_port_scan ) {
    fprintf(stderr,"Program will only be tested in either:\n");
    fprintf(stderr,"TCP SYN Flood detection or Port Scan detection mode\n");
    return 0; }
  
  /* Setup Pcap - packet capture interface */
  /* Define the device */
  dev = pcap_lookupdev(errbuf);
  printf("Listening on %s.\n", dev);
  
  /* Find the properties for the device */
  pcap_lookupnet(dev, &net, &mask, errbuf);
  
  /* Open the session in promiscuous mode */
  handle = pcap_open_live(dev, BUFSIZ, 1, 0, errbuf);
  
  /* You decide when to break */
  while (1) {

    packet = pcap_next(handle, &header);

    /* Get the Ethernet header */
    ether_header = (struct ether_header*)(packet);
    
    /* ---------------------------------------------------------------
     * If we have an IP packet... 
     * --------------------------------------------------------------- */
    if ( ntohs(ether_header->ether_type) == ETHERTYPE_IP ) {
      
      /* Get the IP header */
      ip = (struct ip*)(packet + size_ethernet);     
      
      /* -----------------------------------------------------------
       * we have a TCP packet...
       * ----------------------------------------------------------- */
      if (ip->ip_p == IPPROTO_TCP) {
	
	tcphdr = (struct tcphdr*)(packet + size_ethernet + size_ip);

	dst_port = ntohs(tcphdr->dest);	    
	src_port = ntohs(tcphdr->source);
	
      }
      
      /* -----------------------------------------------------------
       * ICMP pkt...
       * ----------------------------------------------------------- */
      if (ip->ip_p == IPPROTO_ICMP) {
	
	icmp = (struct icmphdr*)(packet + size_ethernet + size_ip);

      }
    }
  }
  pcap_close(handle);
}

/* ------------------------------------------------------------
 *
 *
 *
 * ------------------------------------------------------------ */
static void usage(char* argv0) {

  printf("Intrusion Detection System\n");

  printf("Format: %s [-h] [-t tcp_syn_thresh]   [-s tcp_syn_time]\n",argv0);

  printf("                [-p port_scan_thresh] [-S port_scan_time]\n");

  printf("                [-f tcp_syn_filename] [-F port_scan_filename]\n");
}
